 // Author:      Brent Baccala 
 // Adopted for C++ by  Plyashkevich Viatcheslav <plyashkevich@yandex.ru>

 // This program is free software; you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
 // the Free Software Foundation; either version 2 of the License, or
 // (at your option) any later version.
 //--------------------------------------------------------------------------
 // All rights reserved. 

#include "hdlc.hpp"

struct Hdlc::StcAux
{
	UINT16  state;
	UINT16  r_one;
	UINT16  r_val;
	UINT16  insane_mode;
	UINT32  o_bitcnt;
	UINT32  i_bitcnt;
	UINT32  fcs;
};

static UINT16 fcstab[256] =
{
	0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
	0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
	0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
	0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
	0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
	0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
	0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
	0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
	0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
	0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
	0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
	0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
	0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
	0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
	0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
	0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
	0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
	0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
	0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
	0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
	0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
	0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
	0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
	0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
	0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
	0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
	0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
	0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
	0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
	0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
	0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
	0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78
};
 
#define PPP_INITFCS	0xffff	
#define PPP_GOODFCS	0xf0b8
#define PPP_FCS(fcs, c)	(((fcs) >> 8) ^ fcstab[((fcs) ^ (c)) & 0xff])

#define HDLC_ZERO_SEARCH 0
#define HDLC_FLAG_SEARCH 1
#define HDLC_FLAG_FOUND  2
#define HDLC_FRAME_FOUND 3
#define HDLC_NULL 4
#define HDLC_PART 5
#define HDLC_FULL 6

#define HDLC_FLAG_VALUE	0x7e

int Hdlc::hdlcPack(UINT8 *src, UINT32 slen, UINT8 *dst, UINT32 dsize)
{
	register UINT32 i,d_cnt=0;
	register UINT8 j;
	register UINT8 val;
	register UINT8 s_one = 0;
	register UINT8 out_val = 0;
	register UINT8 bitcnt = 0;
	UINT32 fcs;
	
	if(!dsize)
		return -1;
	
	dst[d_cnt++] = HDLC_FLAG_VALUE;
	fcs = PPP_INITFCS;
	for (i=0; i<slen; i++) {
		val = src[i];
		fcs = PPP_FCS (fcs, val);
		for (j=0; j<8; j++) { 
			bitcnt++;
			out_val >>= 1;
			if (val & 1) {
				s_one++;
				out_val |= 0x80;
			} else {
				s_one = 0;
				out_val &= 0x7f;
			}
			if (bitcnt==8) {
				if (d_cnt == dsize) 
					return 0;
				dst[d_cnt++] = out_val;
				bitcnt = 0;
			}
			if (s_one == 5) {
				out_val >>= 1;
				out_val &= 0x7f;
				bitcnt++;
				s_one = 0;
			}
			if (bitcnt==8) {
				if (d_cnt == dsize) 
					return 0;
				dst[d_cnt++] = out_val;
				bitcnt = 0;
			}
			val >>= 1;
		}
	}
	fcs ^= 0xffff;
	val = fcs & 0xff;
	for (j=0; j<8; j++) { 
			bitcnt++;
			out_val >>= 1;
			if (val & 1) {
				s_one++;
				out_val |= 0x80;
			} else {
				s_one = 0;
				out_val &= 0x7f;
			}
			if (bitcnt==8) {
				if (d_cnt == dsize) 
					return 0;
				dst[d_cnt++] = out_val;
				bitcnt = 0;
			}
			if (s_one == 5) {
				out_val >>= 1;
				out_val &= 0x7f;
				bitcnt++;
				s_one = 0;
			}
			if (bitcnt==8) {
				if (d_cnt == dsize) 
					return 0;
				dst[d_cnt++] = out_val;
				bitcnt = 0;
			}
			val >>= 1;
		}

	val = (fcs>>8) & 0xff;
	for (j=0; j<8; j++) { 
			bitcnt++;
			out_val >>= 1;
			if (val & 1) {
				s_one++;
				out_val |= 0x80;
			} else {
				s_one = 0;
				out_val &= 0x7f;
			}
			if (bitcnt==8) {
				if (d_cnt == dsize) 
					return 0;
				dst[d_cnt++] = out_val;
				bitcnt = 0;
			}
			if (s_one == 5) {
				out_val >>= 1;
				out_val &= 0x7f;
				bitcnt++;
				s_one = 0;
			}
			if (bitcnt==8) {
				if (d_cnt == dsize) 
					return 0;
				dst[d_cnt++] = out_val;
				bitcnt = 0;
			}
			val >>= 1;
		}

	val = HDLC_FLAG_VALUE;
	for (j=0; j<8; j++) { 
		bitcnt++;
		out_val >>= 1;
		if (val & 1)
			out_val |= 0x80;
		else
			out_val &= 0x7f;
		if (bitcnt==8) {
			if (d_cnt == dsize) 
				return 0;
			dst[d_cnt++] = out_val;
			bitcnt = 0;
		}
		val >>= 1;
	}
	if (bitcnt) {
		while (8>bitcnt++) {
			out_val >>= 1;
			out_val |= 0x80;
		}
		if (d_cnt == dsize) 
			return 0;
		dst[d_cnt++] = out_val;
	}

	return d_cnt;
}

int Hdlc::hdlcUnPack(UINT8 *src, UINT32 slen, UINT8 *dst, UINT32 dsize)
{
	INT32 retval=0;
	register UINT8 val;
	register UINT8 state = pStcAux_->state;
	register UINT8 r_one = pStcAux_->r_one;
	register UINT8 r_val = pStcAux_->r_val;
	register UINT32 o_bitcnt = pStcAux_->o_bitcnt;
	register UINT32 i_bitcnt = pStcAux_->i_bitcnt;
	register UINT32 fcs    = pStcAux_->fcs;
	register UINT32 *isrc = (UINT32 *) src;

	if(!dsize)
		return -1;

	if ((i_bitcnt >> 3) < slen) {
		if (pStcAux_->insane_mode==1) {
			val = isrc[(i_bitcnt >> 3)] & 0xff;
		} else if (pStcAux_->insane_mode==2) {
			val = (isrc[i_bitcnt >> 3] >>8) & 0xff;
		} else {
			val = src[i_bitcnt >> 3];
		}
		val >>= i_bitcnt & 7;
	}

	while ((retval == 0) && ((i_bitcnt >> 3) < slen)) {
		if ((i_bitcnt & 7) == 0) {
			if (pStcAux_->insane_mode==1) {
				val = isrc[(i_bitcnt >> 3)] & 0xff;
			} else if (pStcAux_->insane_mode==2) {
				val = (isrc[i_bitcnt >> 3] >>8) & 0xff;
			} else {
				val = src[i_bitcnt >> 3];
			}

			if (val == 0xff) {
				state = HDLC_ZERO_SEARCH;
				o_bitcnt = 0;
				r_one = 0;
				i_bitcnt += 8;
				continue;
			}
		}

		if (state == HDLC_ZERO_SEARCH) {
			if (val & 1) {
				r_one++;
			} else {
				r_one=0;
				state= HDLC_FLAG_SEARCH;
			}
		} else if (state == HDLC_FLAG_SEARCH) { 
			if (val & 1) {
				r_one++;
				if (r_one>6) {
					state=HDLC_ZERO_SEARCH;
				}
			} else {
				if (r_one==6) {
					o_bitcnt=0;
					r_val=0;
					state=HDLC_FLAG_FOUND;
				}
				r_one=0;
			}
		} else if (state ==  HDLC_FLAG_FOUND) {
			if (val & 1) {
				r_one++;
				if (r_one>6) {
					state=HDLC_ZERO_SEARCH;
				} else {
					r_val >>= 1;
					r_val |= 0x80;
					o_bitcnt++;
				}
			} else {
				if (r_one==6) {
					o_bitcnt=0;
					r_val=0;
					r_one=0;
					i_bitcnt++;
					val >>= 1;
					continue;
				} else if (r_one!=5) {
					r_val >>= 1;
					r_val &= 0x7f;
					o_bitcnt++;
				}
				r_one=0;	
			}
			if ((state != HDLC_ZERO_SEARCH) &&
				!(o_bitcnt & 7)) {
				state=HDLC_FRAME_FOUND;
				fcs = PPP_INITFCS;
				dst[0] = r_val;
				fcs = PPP_FCS (fcs, r_val);
			}
		} else if (state ==  HDLC_FRAME_FOUND) {
			if (val & 1) {
				r_one++;
				if (r_one>6) {
					state=HDLC_ZERO_SEARCH;
					o_bitcnt=0;
				} else {
					r_val >>= 1;
					r_val |= 0x80;
					o_bitcnt++;
				}
			} else {
				if (r_one==6) {
					r_val=0; 
					r_one=0;
					o_bitcnt++;
					if (o_bitcnt & 7) {						
						state=HDLC_FLAG_SEARCH;
						retval = -1;
					} else if (fcs==PPP_GOODFCS) {						
						state=HDLC_FLAG_FOUND;
						retval = (o_bitcnt>>3)-3;
					} else {						
						state=HDLC_FLAG_FOUND;
						retval = -1;
					}
				} else if (r_one==5) {
					r_one=0;
					i_bitcnt++;
					val >>= 1;
					continue;
				} else {
					r_val >>= 1;
					r_val &= 0x7f;
					o_bitcnt++;
				}
				r_one=0;	
			}
			if ((state == HDLC_FRAME_FOUND) &&
				!(o_bitcnt & 7)) {
				if ((o_bitcnt>>3)>=dsize) {					
					r_val=0; 
					state=HDLC_FLAG_SEARCH;
					retval = -1;
				} else {
					dst[(o_bitcnt>>3)-1] = r_val;
					fcs = PPP_FCS (fcs, r_val);
				}
			}
		}
		i_bitcnt ++;
		val >>= 1;
	}


	if (retval == 0) {
		i_bitcnt = 0;
	} else {
		o_bitcnt = 0;
	}

	pStcAux_->state = state;
	pStcAux_->r_one = r_one;
	pStcAux_->r_val = r_val;
	pStcAux_->fcs = fcs;
	pStcAux_->o_bitcnt = o_bitcnt;
	pStcAux_->i_bitcnt = i_bitcnt;

	if(retval != 0)
	{
		pStcAux_->state = HDLC_ZERO_SEARCH;
		pStcAux_->r_one = 0;
		pStcAux_->r_val = 0;
		pStcAux_->o_bitcnt = 0;
		pStcAux_->i_bitcnt = 0;
	}
	return (retval);
}

Hdlc::Hdlc(UINT16 mode):pStcAux_(new Hdlc::StcAux)
{
	pStcAux_->state = HDLC_ZERO_SEARCH;
	pStcAux_->r_one = 0;
	pStcAux_->r_val = 0;
	pStcAux_->o_bitcnt = 0;
	pStcAux_->i_bitcnt = 0;
	pStcAux_->insane_mode = mode;
}

Hdlc::~Hdlc()
{
	delete pStcAux_;
}



